unit TTwitterBot2ClassUnit;

{
  by melanion.info

  Media restrictions:
    Image (png, jpeg, gif, webp)
      file Size <= 5 MB

    GIF (animated)
      file Size <= 15 MB
      Resolution <= 1280x1080 (width x height)
      Number of frames <= 350
      Number of pixels (width * height * num_frames) <= 300 million

    Video
      file Size <= 512 MB
      Codec H264 High Profile (Recommended)
      Frame rate <= 60 FPS
      Dimensions 32x32 ... 1280x1024
      Duration 0.5 ... 140 seconds
      Aspect ratio  1:3 ... 3:1
      pixel aspect ratio 1:1
      pixel format Only YUV 4:2:0
      Audio AAC with Low Complexity profile
      Audio mono or stereo
      open GOP
      progressive scan
}

interface

uses
  System.SysUtils, System.Classes, System.DateUtils, System.StrUtils,
  System.Types,

  IdHMACSHA1, IdCoderMIME, IdGlobal,

  System.Net.HttpClient, System.Net.Mime, System.Net.URLClient, System.Hash,

  System.IOUtils, System.JSON, System.NetEncoding, System.Math,
  System.JSON.Serializers, System.JSON.Converters,
  System.Generics.Collections, System.SyncObjs, System.Threading;

type
  TAuthState = (osNoneAuthorized, osInProcess, osAuthorized);
  THTTPProc = (hpNone, hpAUth_v1, hpAuth_v2);
  THTTPMethod = (hmGet, hmPost);
  TAPIVersion = (av11, av20);

  TBrowserReset = procedure () of object;
  TBrowserFormShowEvent = procedure (Url: string; Hint: string) of object;
  TUrlChangedEvent = procedure (NewUrl: string; var DoCloseWebForm : boolean) of object;
  TResultEvent = reference to procedure (const AResult: string; const AuthStateChanged: boolean);
  TAuthStateChangedEvent = procedure (AuthState: TAuthState; APIVersion: TAPIVersion) of object;
  TSetJsonEvent = procedure (Json: string) of object;

  TMessageRec = record
    Name: string;
    Text: string;
  end;

  TTwitterBot2 = class
  strict private
    // API v1.1
    v1_ConsumerKey: string;
    v1_ConsumerKeySecret: string;
    v1_OAuthtoken: string;
    v1_OAuthtokenSecret: string;
    v1_RefreshToken: string;
    v1_AccessToken: string;          // obtained by auth
    v1_AccessTokenSecret: string;    // obtained by auth
    v1_AuthState: TAuthState;

    // API v2
    v2_ClientID: string;
    v2_ClientSecret: string;
    v2_Scopes: string;
    v2_RedirectURL: string;
    v2_ChallengeMethod: string;
    v2_CodeVerifier: string;
    v2_Challenge: string;
    v2_RefreshToken: string;         // obtained by auth
    v2_BearerToken: string;          // obtained by auth
    v2_RefreshedAt: TDateTime;
    v2_AuthState: TAuthState;
    v2_State: string;

    FHTTPProc: THTTPProc;
    FAuthStateChanged: TAuthStateChangedEvent;

    FBrowserFormShow: TBrowserFormShowEvent;

    FSetJsonEvent_v1: TSetJsonEvent;
    FSetJsonEvent_v2: TSetJsonEvent;

    FRateLimit: string;

    procedure LogRateLimit(Header: TNetHeaders);

    // v1
    procedure Set_v1_Tokens(Json: string);
    function Get_v1_TokensAsJson: string;
    // upload a image file
    function DoUploadImage_v1(Filename: string): string;
    // upload media
    function DoChunkedUploadMedia_v1(Filename: string): string;
    function DoUploadMediaAppend_v1(Filename, MediaID, MediaType, Signature: string): boolean;
    procedure AsyncHTTP_v1(URL, Signature, MediaID: string; SegIndex: integer; Stream: TMemoryStream);
    function DoUploadMediaStatus_v1(MediaID: string; WaitSeconds: integer): boolean;

    // v2
    procedure Set_v2_Tokens(Json: string);
    function Get_v2_TokensAsJson: string;
    function ParseTokensFromOrgJson(var Json: string; DateTime: TDateTime): boolean;
    // refresh an expired access token (a access token expires after 7200 second)
    function DoRefreshTokens_v2: string;
    // post text with MediaID of Image
    function DoPostTweet_v2(Text: string; MediaID: string): string;
    // get /users/me
    function DoGetAboutMe: string;
  public
    // upload larger media file (mp4 or animated GIF) async and pararel (simultaneously)
    AsyncChunkedMediaUpload: boolean;

    constructor Create;
    destructor Destroy; override;

    function CreateSignature(Method: THTTPMethod; URL, OptionParams: string): string; // result is OAuth string

    // events
    //property UrlChanged: TUrlChangedEvent write FUrlChanged;                    // to MainForm
    property AuthStateChanged: TAuthStateChangedEvent write FAuthStateChanged;  // to MainForm

    // v1
    property v1_Tokens: string read Get_v1_TokensAsJson write Set_v1_Tokens;
    property SetJsonEvent_v1: TSetJsonEvent write FSetJsonEvent_v1;
    property AuthState_v1: TAuthState read v1_AuthState;
    // v1 authentication starts
    procedure Authorize_v1;                                                     // v1 Auth
    // get an access token by PIN
    procedure GetAccessToken_v1(Verifier: string);                              // v1 Auth

    // v2
    property v2_Tokens: string read Get_v2_TokensAsJson write Set_v2_Tokens;
    property SetJsonEvent_v2: TSetJsonEvent write FSetJsonEvent_v2;
    property AuthState_v2: TAuthState read v2_AuthState;
    // v2 authentication starts
    procedure Authorize_v2;                                                     // v2 Auth
    // get an acerss token and a refresh token by the Code including in URL
    function GetAccessToken_v2(Url: string): string;

    // Post a tweet with media
    procedure PostTweet(Text: string; FileName: string; IsAnimatedGif: boolean; AResultEvent: TResultEvent); // Text: v2 - media: v1

    procedure GetAboutMe(AResultEvent: TResultEvent);

    // tools
    property BrowserFormShow: TBrowserFormShowEvent write FBrowserFormShow;
    procedure BrowserNavigate(URL: string);

    property RateLimit: string read FRateLimit;

  end;

  function GetMimeStr(Filename: string): string;

  var
    TwitterBot2: TTwitterBot2 = nil;

    AppClosing: boolean = false;

    LogQueue: TThreadList<string> = nil;

implementation



const
  // v1
  RequestTokenURL = 'https://api.twitter.com/oauth/request_token';
  RequestVerifierURL = 'https://twitter.com/oauth/authorize?oauth_token=';
  RequestAccessTokenURL = 'https://api.twitter.com/oauth/access_token';
  UploadMediaURL_v1 = 'https://upload.twitter.com/1.1/media/upload.json';
  PostTweetURL_v1 = 'https://api.twitter.com/1.1/statuses/update.json';
  // v2
  AuthorizeURL = 'https://twitter.com/i/oauth2/authorize';
  GetAccessTokenURL = 'https://api.twitter.com/2/oauth2/token';
  UploadMediaUrl_v2 = 'https://upload.twitter.com/1.1/media/upload.json';
  PostTweetUrl_v2 = 'https://api.twitter.com/2/tweets';
  AboutMeUrl = 'https://api.twitter.com/2/users/me';

  ChunkSize = 4 * 1024 * 1024; // 4MB Media
  Size5MB =  5 * 1024 * 1024; // 5MB Media


// percent encode
function UrlEncode(Text: string): string;
const
  UrlEncodeCharSet = ['A'..'Z', 'a'..'z', '0'..'9', '-', '.', '_', '~'];
var
  Ch: Char;
begin
  result := '';

  try
    for ch in Text do
      if CharInSet(Ch, UrlEncodeCharSet) then
        result := result + Char(Ch)
      else
        result := result + '%'+ IntToHex(ord(Ch), 2);
  except

  end;
end;

// add a Key and Value pair to a Parameter string (for header)
procedure AddParam(var Param: string; Key, Value: string);
begin
  if Param = '' then
    Param := Key + '=' + '"' + Value + '"'
  else
    Param := Param + ',' + Key + '=' + '"' + Value + '"';
end;

procedure AddQueryStr(var Param: string; Key, Value: string);
begin
  if Param = '' then
    Param := Key + '=' + Value
  else
    Param := Param + '&' + Key + '=' + Value;
end;

function MemoryStreamToString(Stream: TMemoryStream): string;
var
  SList: TStringList;
begin
  result := '';
  if Assigned(Stream) and (Stream.size > 0) then begin
    SList := TStringList.Create;
    try
      try
        Stream.Seek(0, TSeekOrigin.soBeginning);
        SList.LoadFromStream(Stream, TEncoding.UTF8);
        result := SList.Text;
      except

      end;
    finally
      SList.Free;
    end;
  end;
end;

{ TTwitterBot2 }

constructor TTwitterBot2.Create;
begin
  FBrowserFormShow := nil;
  FHTTPProc := THTTPProc.hpNone;

  //FUrlChanged := nil;
  FAuthStateChanged := nil;
  FSetJsonEvent_v1 := nil;
  FSetJsonEvent_v2 := nil;

  FRateLimit := '';

  AsyncChunkedMediaUpload := false;
end;

destructor TTwitterBot2.Destroy;
begin

  inherited;
end;

procedure TTwitterBot2.Set_v1_Tokens(Json: string);
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        try
          if JSONObject.TryGetValue('consumer_key', JsonValue) then
            v1_ConsumerKey := JsonValue.Value;
          if JSONObject.TryGetValue('consumer_key_secret', JsonValue) then
            v1_ConsumerKeySecret := JsonValue.Value;

          if JSONObject.TryGetValue('bearer_token', JsonValue) then
            v1_RefreshToken := JsonValue.Value;

          if JSONObject.TryGetValue('access_token', JsonValue) then begin
            v1_AccessToken := JsonValue.Value;
            v1_OAuthToken := v1_AccessToken;
          end;
          if JSONObject.TryGetValue('access_token_secret', JsonValue) then begin
            v1_AccessTokenSecret := JsonValue.Value;
            v1_OAuthTokenSecret := v1_AccessTokenSecret;
          end;

          if not v1_AccessToken.IsEmpty and not v1_AccessTokenSecret.IsEmpty then
            v1_AuthState := TAuthState.osAuthorized
          else
            v1_AuthState := TAuthState.osNoneAuthorized;
          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
        except
          on e: exception do
            if Assigned(LogQueue) then
              LogQueue.Add('TTwitterBot2.Set_v1_Tokens: ' + e.Message);
        end;
      finally
        JSONObject.Free;
      end;
    end;
  end;
end;

function TTwitterBot2.Get_v1_TokensAsJson: string;
var
  JSON: TJSONObject;
begin
  result := '';

  JSON := TJSONObject.Create;
  try
    try
      JSON.AddPair('consumer_key', v1_ConsumerKey);
      JSON.AddPair('consumer_key_secret', v1_ConsumerKeySecret);

      JSON.AddPair('bearer_token', v1_RefreshToken);

      JSON.AddPair('access_token', v1_AccessToken);
      JSON.AddPair('access_token_secret', v1_AccessTokenSecret);
      result := Json.ToString;
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.Get_v1_TokensAsJson: ' + e.Message);
    end;
  finally
    JSON.Free;
  end;
end;

procedure TTwitterBot2.Set_v2_Tokens(Json: string);
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        try
          if JSONObject.TryGetValue('client_id', JsonValue) then
            v2_ClientID := JsonValue.Value;
          if JSONObject.TryGetValue('client_id_secret', JsonValue) then
            v2_ClientSecret := JsonValue.Value;

          if JSONObject.TryGetValue('scopes', JsonValue) then
            v2_Scopes := JsonValue.Value;
          if JSONObject.TryGetValue('redirect_url', JsonValue) then
            v2_RedirectURL := JsonValue.Value;
          if JSONObject.TryGetValue('challenge_method', JsonValue) then
            v2_ChallengeMethod := JsonValue.Value;

          if JSONObject.TryGetValue('refresh_token', JsonValue) then
            v2_RefreshToken := JsonValue.Value;
          if JSONObject.TryGetValue('bearer_token', JsonValue) then
            v2_BearerToken := JsonValue.Value;
          v2_RefreshedAt := 0;
          if JSONObject.TryGetValue('bearer_token_obtained_at', JsonValue) then begin
            if not JsonValue.Value.IsEmpty then
              v2_RefreshedAt := ISO8601ToDate(JsonValue.Value, false);
          end;

          if not v2_RefreshToken.IsEmpty and not v2_BearerToken.IsEmpty then
            v2_AuthState := TAuthState.osAuthorized
          else
            v2_AuthState := TAuthState.osNoneAuthorized;

          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v2_AuthState, TAPIVersion.av20);
        except
          on e: exception do
            if Assigned(LogQueue) then
              LogQueue.Add('TTwitterBot2.Set_v2_Tokens: ' + e.Message);
        end;
      finally
        JSONObject.Free;
      end;
    end;
  end;

end;

function TTwitterBot2.Get_v2_TokensAsJson: string;
var
  JSON: TJSONObject;
begin
  result := '';

  JSON := TJSONObject.Create;
  try
    try
      JSON.AddPair('client_id', v2_ClientID);
      JSON.AddPair('client_id_secret', v2_ClientSecret);

      JSON.AddPair('scopes', v2_Scopes);
      JSON.AddPair('redirect_url', v2_RedirectURL);
      JSON.AddPair('challenge_method', v2_ChallengeMethod);

      JSON.AddPair('refresh_token', v2_RefreshToken);
      JSON.AddPair('bearer_token', v2_BearerToken);
      JSON.AddPair('bearer_token_obtained_at', DateToISO8601(v2_RefreshedAt, false));

      result := Json.ToString;
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.Get_v2_TokensAsJson: ' + e.Message);
    end;
  finally
    JSON.Free;
  end;

end;

{
procedure TTwitterBot2.SaveTokens;
var
  AText: string;
begin
   var AIni := TMemIniFile.Create(ExtractFileDir(ParamStr(0)) + PathDelim + 'Tokens.ini', TEncoding.UTF8);
   try
     try
       // v1
       AIni.WriteString('v1_Tokens', 'ConsumerKey', v1_ConsumerKey);
       AText := Encrypt(v1_ConsumerKeySecret);
       AIni.WriteString('v1_Tokens', 'ConsumerKeySecret', AText);
       AIni.WriteString('v1_Tokens', 'AccessToken', v1_AccessToken);
       AText := Encrypt(v1_AccessTokenSecret);
       AIni.WriteString('v1_Tokens', 'AccessTokenSecret', AText);

       // v2
       AIni.WriteString('v2_Tokens', 'ClientID', v2_ClientID);
       AText := Encrypt(v2_ClientSecret);
       AIni.WriteString('v2_Tokens', 'ClientSecret', AText);
       AText := Encrypt(v2_RefreshToken);
       AIni.WriteString('v2_Tokens', 'RefreshToken', AText);
       AText := Encrypt(v2_BearerToken);
       AIni.WriteString('v2_Tokens', 'BearerToken', AText);
       AText := DateToISO8601(v2_RefreshedAt, false);
       AIni.WriteString('v2_Tokens', 'BearerObtainedAt', AText);

       AIni.UpdateFile;
     except

     end;
   finally
     AIni.Free;
   end;
end;}

function CreateHttpClient: THTTPClient;
begin
  result := THTTPClient.Create;

  result.AllowCookies := true;
  result.MaxRedirects := 10;
  result.HandleRedirects := true;

  result.Accept := 'text/html\tapplication/xhtml+xml\tapplication/xml;q=0.9\timage/webp\timage/apng\t*/*;q=0.8\tapplication/signed-exchange;v=b3;q=0.9';
  result.AcceptLanguage := 'ja\ten-US;q=0.9\ten;q=0.8';
  result.UserAgent := 'TTwitterBot v2.0 by melanion';

  {$IFDEF MSWINDOWS}
  result.AcceptEncoding := 'gzip deflate';
  if result.AcceptEncoding.Contains('gzip') then
    result.AutomaticDecompression := [THTTPCompressionMethod.GZip];
  if result.AcceptEncoding.Contains('deflate') then
    result.AutomaticDecompression := result.AutomaticDecompression + [THTTPCompressionMethod.Deflate];
  {$ENDIF}

  result.ConnectionTimeout := 20000;
  result.ResponseTimeout := 20000;
  result.SendTimeout := 20000;
end;

// RateLimit
procedure TTwitterBot2.LogRateLimit(Header: TNetHeaders);
begin
  if Length(Header) > 0 then begin
    try
      FRateLimit := '';
      for var Pair in Header do begin
        if Pair.Name.Contains('x') and Pair.Name.Contains('-limit-') then begin
          if Pair.Name.Contains('-reset') then begin
            // Unix time
            if not Trim(Pair.Value).IsEmpty then
              FRateLimit := FRateLimit + Pair.Name + '=' +
                  DateToISO8601(UnixToDateTime(Trim(Pair.Value).ToInt64)) + sLineBreak;
          end else
            FRateLimit := FRateLimit + Pair.Name + '=' + Trim(Pair.Value) + sLineBreak;
        end;
      end;
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.LogRateLimit: ' + e.Message);
    end;

   {$IFDEF DEBUG}
    if Assigned(LogQueue) and not FRateLimit.IsEmpty then
      LogQueue.Add(FRateLimit);
   {$ENDIF}
  end;
end;

// Misc

// public
procedure TTwitterBot2.BrowserNavigate(URL: string);
begin
  if Assigned(FBrowserFormShow) then
    FBrowserFormShow(Url, '');

end;

// post a tweet with and image / video
procedure TTwitterBot2.PostTweet(Text: string; FileName: string; IsAnimatedGif: boolean; AResultEvent: TResultEvent);
begin
  FHTTPProc := THTTPProc.hpAuth_v2;

  if not Text.IsEmpty and Assigned(AResultEvent) then begin
    TThread.CreateAnonymousThread(procedure
    var
      MediaID, Response: string;
      Refreshed: boolean;
    begin
      try
        Refreshed := false;
        if IncSecond(v2_RefreshedAt, 7000) < Now then begin // 7200
          if not DoRefreshTokens_v2.IsEmpty then
          Refreshed := true;
        end;

        // Upload an image / a video v1
        if (v1_AuthState = TAuthState.osAuthorized) and not FileName.IsEmpty then begin
          if (TPath.GetExtension(Filename) = '.mp4') or
             (IsAnimatedGif and (TFile.GetSize(Filename) > Size5MB)) then
            MediaID := DoChunkedUploadMedia_v1(FileName)
          else
            MediaID := DoUploadImage_v1(FileName)
        end else
          MediaID := '';

        // Post a tweet v2
        if not Text.IsEmpty then
          Response := DoPostTweet_v2(Text, MediaID)
        else
          Response := MediaID;

        TThread.Synchronize(
          TThread.CurrentThread,
          procedure
          begin
            AResultEvent(Response, false);
            if Assigned(FSetJsonEvent_v2) and Refreshed then
              FSetJsonEvent_v2(Get_v2_TokensAsJson);
          end);
      except
        on e: exception do
          if Assigned(LogQueue) then
            LogQueue.Add('TTwitterBot2.PostTweet: ' + e.Message);
      end;
    end).Start;
  end;

end;

// get my account info
procedure TTwitterBot2.GetAboutMe(AResultEvent: TResultEvent);
begin
  if Assigned(AResultEvent) then begin
    TThread.CreateAnonymousThread(procedure
    var
      Json, AResult: string;
      Refreshed: boolean;
      JSONObject: TJSONObject;
      JsonValue: TJsonValue;
    begin
      try
        Refreshed := false;
        if IncSecond(v2_RefreshedAt, 7000) < Now then begin // 7200
          if not DoRefreshTokens_v2.IsEmpty then
            Refreshed := true;
        end;

        // temp
        Json := DoGetAboutMe;

        AResult := '';
        if Json.Contains('data') then begin
          JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
          if JSONObject <> nil then begin
            try
              if JSONObject.TryGetValue('data.id', JsonValue) then
                AResult := 'id=' + JsonValue.Value;
              if JSONObject.TryGetValue('data.name', JsonValue) then
                AResult := AResult + ', name=' + JsonValue.Value;
              if JSONObject.TryGetValue('data.username', JsonValue) then
                AResult := AResult + ', username=' + JsonValue.Value;
            finally
              JSONObject.Free;
            end;
          end;

          TThread.Synchronize(
            TThread.CurrentThread,
            procedure
            begin
              AResultEvent(AResult, false);
              if Assigned(FSetJsonEvent_v2) and Refreshed then
                FSetJsonEvent_v2(Get_v2_TokensAsJson);
            end);
        end;
      except
        on e: exception do
          if Assigned(LogQueue) then
            LogQueue.Add('TTwitterBot2.GetAboutMe: ' + e.Message);
      end;
    end).Start;
  end;

end;



// ---------------------------------------------
//     v1
// ---------------------------------------------

// result is OAuth query string
function TTwitterBot2.CreateSignature(Method: THTTPMethod; URL, OptionParams: string): string;
  // generate 32 bytes Nonce
  function CreateNonce: string;
  var
    i: integer;
  const
    Str: string = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
  begin
    Randomize;
    for i := 0 to 31 do
      result := result + Str[Random(62) + 1];
  end;
  // calculate the signature by SHA1
  function CalcHMACSha1(SignatureBase, SigningKey: string): string;
  var
    SHA1: TIdHMACSHA1;
  begin
    result := '';
    SHA1 := TIdHMACSHA1.Create;
    try
      SHA1.Key := ToBytes(SigningKey);
      result := TIdEncoderMIME.EncodeBytes(SHA1.HashValue(ToBytes(SignatureBase)));
    finally
      SHA1.Free;
    end;
  end;
const
  HTTPMethodStr: array[THTTPMethod] of string = ('GET', 'POST');
var
  Nonce: string;
  TimeStamp: string;
  PList: TStringList;
  SArray: TStringDynArray;
  Params: string;
  SignatureKey: string;
  SignatureBase: string;
  Signature: string;
begin
  result := '';

  PList := TStringList.Create;
  try
    try
      // calc signatur
      Nonce := CreateNonce;
      TimeStamp := IntToStr(DateTimeToUnix(Now, true));

      PList.Add('oauth_consumer_key=' + URLEncode(v1_ConsumerKey));
      PList.Add('oauth_nonce=' + Nonce);
      PList.Add('oauth_signature_method=HMAC-SHA1');
      PList.Add('oauth_timestamp=' + TimeStamp);
      if not v1_OAuthToken.IsEmpty then
        PList.Add('oauth_token=' + URLEncode(v1_OAuthToken));
      PList.Add('oauth_version=1.0');
      if not OptionParams.IsEmpty then begin
        SArray := SplitString(OptionParams, '&');
        for var Text in SArray do
          PList.Add(Text);
      end;

      // sort
      PList.Sort;

      Params := '';
      for var Text in PList do
        if Params.IsEmpty then
          Params := Text
        else
          Params := Params + '&' + Text;

      SignatureKey := UrlEncode(v1_ConsumerKeySecret) + '&' + UrlEncode(v1_OAuthTokenSecret);
      SignatureBase := HTTPMethodStr[Method] + '&' + UrlEncode(URL) + '&' + UrlEncode(Params);
      Signature := UrlEncode(CalcHMACSha1(SignatureBase, SignatureKey));

      // result = OAuth Header string
      PList.Clear;
      PList.Add('oauth_consumer_key="' + URLEncode(v1_ConsumerKey) + '"');
      PList.Add('oauth_nonce="' + Nonce + '"');
      PList.Add('oauth_signature="' + Signature + '"');
      PList.Add('oauth_signature_method="HMAC-SHA1"');
      PList.Add('oauth_timestamp="' + TimeStamp + '"');
      if not v1_OAuthToken.IsEmpty then
        PList.Add('oauth_token="' + URLEncode(v1_OAuthToken) + '"');
      PList.Add('oauth_version="1.0"');
      for var Text in SArray do  // add OptionParams that start with 'oauth_'
        if Text.StartsWith('oauth_') then begin
          var QArray := SplitString(Text, '=');
          if Length(QArray) = 2 then
            PList.Add(QArray[0] + '="' + QArray[1] + '"');
        end;
      for var Text in PList do
        if result.IsEmpty then
          result := Text
        else
          result := result + ',' + Text;

      result := 'OAuth ' + result;

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.CreateSignature: ' + e.Message);
    end;
  finally
    PList.Free;
  end;
end;

// obtain a Request Token
// https://developer.twitter.com/en/docs/authentication/api-reference/request_token
procedure TTwitterBot2.Authorize_v1;
var
  HttpClient: THttpClient;
  Body: TStringList;
  Response: TMemoryStream;
  Json: string;
begin
  FHTTPProc := THTTPProc.hpAuth_v1;

  HttpClient := CreateHTTPClient;
  Body := TStringList.Create;
  Response := TMemoryStream.Create;
  try
    try
      HttpClient.CustHeaders.Add('Authorization', CreateSignature(hmPost, RequestTokenURL, 'oauth_callback=oob'));

      HttpClient.Post(RequestTokenURL, Body, Response);

      // parse Response
      Json := MemoryStreamToString(Response);
      if not Json.IsEmpty then begin
        {oauth_token=Z6eEdO8MOmk394WozF5oKyuAv855l4Mlqo7hhlSLik&
         oauth_token_secret=Kd75W4OQfb2oJTV0vzGzeXftVAwgMnEK9MumzYcM&
         oauth_callback_confirmed=true}
        var ResponseList := TStringList.Create;
        try
          ResponseList.Text := Json.Replace('&', sLineBreak);
          v1_OAuthToken := ResponseList.Values['oauth_token'];;
          v1_OAuthTokenSecret := ResponseList.Values['oauth_token_secret'];
        finally
          ResponseList.Free;
        end;
      end;

      if v1_OAuthToken.IsEmpty or v1_OAuthTokenSecret.IsEmpty then begin
        v1_OAuthToken := '';
        v1_OAuthTokenSecret := '';
      end else begin
        if Assigned(FBrowserFormShow) then  // open Browser
          FBrowserFormShow(RequestVerifierURL + v1_OAuthToken, 'Login and authorize a app');
        v1_AuthState := TAuthState.osInProcess;
        if Assigned(FAuthStateChanged) then
          FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
      end;

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.Authorize_v1: ' + e.Message);
    end;
  finally
    HttpClient.Free;
    Body.Free;
    Response.Free;
  end;

end;

// https://developer.twitter.com/en/docs/authentication/api-reference/access_token
procedure TTwitterBot2.GetAccessToken_v1(Verifier: string);
var
  Params: string;
  HttpClient: THttpClient;
  Body: TStringList;
  Response: TMemoryStream;
  Json: string;
begin


  HttpClient := CreateHTTPClient;
  Body := TStringList.Create;
  Response := TMemoryStream.Create;
  try
    try
      Params := '';
      AddQueryStr(Params, 'oauth_token', URLEncode(v1_OAuthToken));
      AddQueryStr(Params, 'oauth_verifier', Verifier);
      HttpClient.CustHeaders.Add('Authorization', CreateSignature(hmPost, RequestAccessTokenURL, Params));

      Body.Add('oauth_verifier=' + Verifier);

      HttpClient.Post(RequestAccessTokenURL, Body, Response);

      // parse Response
      Json := MemoryStreamToString(Response);
      if not Json.IsEmpty then begin
        {oauth_token=6253282-eWudHldSbIaelX7swmsiHImEL4KinwaGloHANdrY&
         oauth_token_secret=2EEfA6BG5ly3sR3XjE0IBSnlQu4ZrUzPiYTmrkVU&
         user_id=6253282&
         screen_name=twitterapi}
        var ResponseList := TStringList.Create;
        try
          ResponseList.Text := Json.Replace('&', sLineBreak);
          v1_OAuthToken := ResponseList.Values['oauth_token'];;
          v1_OAuthTokenSecret := ResponseList.Values['oauth_token_secret'];
        finally
          ResponseList.Free;
        end;
      end;

     if not v1_OAuthToken.IsEmpty and not v1_OAuthTokenSecret.IsEmpty then begin
      v1_AccessToken := v1_OAuthToken;
      v1_AccessTokenSecret := v1_OAuthTokenSecret;
      if Assigned(FBrowserFormShow) then // close Browser
        FBrowserFormShow('', '');
      v1_AuthState := TAuthState.osAuthorized;
      if Assigned(FAuthStateChanged) then
        FAuthStateChanged(v1_AuthState, TAPIVersion.av11);
      if Assigned(FSetJsonEvent_v1) then
        FSetJsonEvent_v1(Get_v1_TokensAsJson);
      end;

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.GetAccessToken_v1: ' + e.Message);
    end;
  finally
    HttpClient.Free;
    Body.Free;
    Response.Free;
  end;

end;

function GetJsonValueByName(Json: string; AName: string): string;
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
 result := '';
 if not Json.IsEmpty and not AName.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        if JSONObject.TryGetValue(AName, JsonValue) then
            result := JsonValue.Value;
      finally
        JSONObject.Free;
      end;
    end;
  end;
end;

function GetTextFromStream(Stream: TMemoryStream): string;
begin
  result := '';
  if Stream.Size > 0 then begin
    Stream.Seek(0, TSeekOrigin.soBeginning);
    var SList := TstringList.Create;
    try
      try
        SList.LoadFromStream(Stream, TEncoding.UTF8);
        result := SList.Text;
      except

      end;
    finally
      SList.Free;
    end;
  end;
end;

// https://developer.twitter.com/ja/docs/media/upload-media/api-reference/post-media-upload
function TTwitterBot2.DoUploadImage_v1(Filename: string): string;
var
  HttpClient: THttpClient;
  Body: TMultipartFormData;
  Stream: TMemoryStream;
  SList: TStringList;
begin
  result := '';

  if not TFile.Exists(Filename) then
    exit;

  HttpClient := CreateHttpClient;
  Body := TMultipartFormData.Create;
  Stream := TMemoryStream.Create;
  try
    try
      HttpClient.ContentType := 'multipart/form-data';
      HttpClient.CustHeaders.Add('Authorization', CreateSignature(hmPost, UploadMediaURL_v1, ''));

      Body.AddFile('media', Filename, 'application/octet-stream');

      HttpClient.Post(UploadMediaURL_v1, Body, Stream);

      if Stream.Size > 0 then begin
        Stream.Seek(0, TSeekOrigin.soBeginning);
        SList := TstringList.Create;
        try
          SList.LoadFromStream(Stream, TEncoding.UTF8);
          result := GetJsonValueByName(SList.Text, 'media_id_string');
        finally
          SList.Free;
        end;
      end;

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.UploadImage: ' + e.Message);
    end;
  finally
    HttpClient.Free;
    Body.Free;
    Stream.Free;
  end;

end;

function GetMimeStr(Filename: string): string;
const
  MimeStr: array[0..5] of string = ('.png:image/png', '.jpg:image/jpeg', '.jpeg:image/jpeg',
                                    '.gif:image/gif', '.mp4:video/mp4', '.webp:image/webp');
begin
  result := '';

  var Ext := LowerCase(TPath.GetExtension(Filename));
  if not Ext.IsEmpty then
    for var i := 0 to High(MimeStr) do begin
      if MimeStr[i].StartsWith(Ext) then begin
        var SArray := SplitString(MimeStr[i], ':');
        if Length(SArray) = 2 then begin
          result := SArray[1];
          Exit;
        end;
      end;
    end;
end;

function GetWaitSecond(Json: string): integer;
begin
  result := 0;
  if Json.Contains('check_after_secs') then begin
   var s := GetJsonValueByName(Json, 'processing_info.check_after_secs');
    TryStrToInt(s, result);
  end;
end;

// https://developer.twitter.com/en/docs/twitter-api/v1/media/upload-media/uploading-media/media-best-practices
function TTwitterBot2.DoChunkedUploadMedia_v1(Filename: string): string;
var
  FileSize: Int64;
  Params: string;
  MediaType: string;
  HttpClient: THttpClient;
  Body: TMultipartFormData;
  Response: TMemoryStream;
  Json: string;
  MediaID: string;
  WaitSeconds: integer;
begin
  result := '';

  if not TFile.Exists(Filename) then
    exit;

  FileSize := TFile.GetSize(Filename);
  if FileSize = 0 then
    Exit;

  MediaType := GetMimeStr(Filename);
  if MediaType.IsEmpty then
    Exit;

  HttpClient := CreateHttpClient;
  Body := TMultipartFormData.Create;
  Response := TMemoryStream.Create;
  try
    try
      HttpClient.ContentType := 'multipart/form-data';
      Params := CreateSignature(hmPost, UploadMediaURL_v1, '');
      HttpClient.CustHeaders.Add('Authorization', Params);

      // init
      Body.AddField('command', 'INIT');
      Body.AddField('total_bytes', FileSize.ToString);
      Body.AddField('media_type', MediaType);
      if MediaType = 'video/mp4' then
        Body.AddField('media_category', 'tweet_video')
      else if MediaType = 'image/gif' then
        Body.AddField('media_category', 'tweet_gif');

      HttpClient.Post(UploadMediaURL_v1, Body, Response);
      Response.Seek(0, TSeekOrigin.soBeginning);
      Json := GetTextFromStream(Response);
      if Json.IsEmpty then
        Exit;
      MediaID := GetJsonValueByName(Json, 'media_id_string');
      if MediaID.IsEmpty then
        Exit;

      if Assigned(LogQueue) then
        LogQueue.Add('DoChunkedUploadMedia_v1-INIT: Done');

      // Append
      if not DoUploadMediaAppend_v1(Filename, MediaID, MediaType, Params) then
        Exit;

      // Finalize
      Response.Clear;
      Body.Free;
      Body := TMultipartFormData.Create;
      Body.AddField('command', 'FINALIZE');
      Body.AddField('media_id', MediaID);
      HttpClient.Post(UploadMediaURL_v1, Body, Response);
      Response.Seek(0, TSeekOrigin.soBeginning);
      Json := GetTextFromStream(Response);
      if Json.IsEmpty then
        Exit
      else
        result := MediaID;
      WaitSeconds := GetWaitSecond(Json);

      if Assigned(LogQueue) then
        LogQueue.Add('DoChunkedUploadMedia_v1-FINALIZE: Done');

      // Status
      if WaitSeconds > 0 then begin
        if not DoUploadMediaStatus_v1(MediaID, WaitSeconds) then
          result := '';
      end;

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoChunkedUploadMedia_v1: ' + e.Message);
    end;
  finally
    HttpClient.Free;
    Body.Free;
    Response.Free;
  end;
end;

function SleepNow(Seconds: integer): boolean;
begin
  result := false;
  if Seconds > 0 then begin
    for var i := 1 to Seconds * 10 do begin
      Sleep(100);
      if AppClosing then
        exit;
    end;
  end;
  result := true;
end;

var
  AsyncHTTPCount: Int64 = 0; // TInterlocked
  AsyncHTTPErrorCount: Int64 = 0;

procedure TTwitterBot2.AsyncHTTP_v1(URL, Signature, MediaID: string; SegIndex: integer; Stream: TMemoryStream);
begin
  if not URL.IsEmpty then begin
    TInterlocked.Increment(AsyncHTTPCount);
    TThread.CreateAnonymousThread(procedure
    var
      HttpClient: THttpClient;
      Body: TMultipartFormData;
      Response: TMemoryStream;
    begin
      HttpClient := CreateHttpClient;
      Body := TMultipartFormData.Create;
      Response := TMemoryStream.Create;
      try
        try
          HttpClient.ContentType := 'multipart/form-data';
          HttpClient.CustHeaders.Add('Authorization', Signature);
          Body.AddField('command', 'APPEND');
          Body.AddField('media_id', MediaID);
          Body.AddField('segment_index', SegIndex.ToString);
          Body.AddStream('media', Stream, '', 'application/octet-stream');

          HttpClient.Post(URL, Body, Response);

          FreeAndNil(Stream);

          TInterlocked.Decrement(AsyncHTTPCount)

        except
          on e: exception do
            TInterlocked.Increment(AsyncHTTPErrorCount);
        end;
      finally
        HttpClient.Free;
        Body.Free;
        Response.Free;
      end;
    end).Start;

  end;
end;
// append chunked stream
function TTwitterBot2.DoUploadMediaAppend_v1(Filename, MediaID, MediaType, Signature: string): boolean;
var
  HttpClient: THttpClient;
  Body: TMultipartFormData;
  Response: TMemoryStream;
  Stream: TMemoryStream;
  FileStream: TFileStream;
begin
  result := false;

  FileStream := TFileStream.Create(Filename, fmOpenRead);
  Stream := TMemoryStream.Create;
  HttpClient := CreateHttpClient;
  Body := TMultipartFormData.Create;
  Response := TMemoryStream.Create;
  var AsyncUpload := AsyncChunkedMediaUpload and (FileStream.Size > ChunkSize);
  try
    try
      FileStream.Seek(0, TSeekOrigin.soBeginning);
      var SegIndex := 0;
      AsyncHTTPCount := 0;
      AsyncHTTPErrorCount := 0;
      while FileStream.Position < FileStream.Size do begin
        if TInterlocked.Read(AsyncHTTPErrorCount) > 0 then Exit;
        if AsyncUpload then begin
          // async and simultaneous upload
          var AStream := TMemoryStream.Create;
          AStream.CopyFrom(FileStream, Min(ChunkSize, FileStream.Size - FileStream.Position));
          AStream.Seek(0, TSeekOrigin.soBeginning);
          AsyncHTTP_v1(UploadMediaURL_v1, Signature, MediaID, SegIndex, AStream);
        end else begin
          // sync upload
          Stream.Clear;
          Stream.CopyFrom(FileStream, Min(ChunkSize, FileStream.Size - FileStream.Position));
          Stream.Seek(0, TSeekOrigin.soBeginning);
          Body.Free;
          Body := TMultipartFormData.Create;
          Body.AddField('command', 'APPEND');
          Body.AddField('media_id', MediaID);
          Body.AddField('segment_index', SegIndex.ToString);
          Body.AddStream('media', Stream, '', 'application/octet-stream');
          Response.Clear;
          HttpClient.Post(UploadMediaURL_v1, Body, Response);
        end;
        Inc(SegIndex);
      end;
      // if AsyncUpload, wait for all chunked uploads finish
      if AsyncUpload then begin
        // wait
        var LoopStartTime := now;
        while TInterlocked.Read(AsyncHTTPCount) > 0 do begin
          if not SleepNow(1) then exit;
          if TInterlocked.Read(AsyncHTTPErrorCount) > 0 then Exit;
          if MinutesBetween(LoopStartTime, Now) > 4 then Exit;    // max wait time = 4 min
        end;
      end;
      result := true;
      if Assigned(LogQueue) then
        LogQueue.Add('DoChunkedUploadMedia_v1-APPEND: Done');
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoUploadMediaAppend_v1: ' + e.Message);
    end;
  finally
    FileStream.Free;
    Stream.Free;
    HttpClient.Free;
    Body.Free;
    Response.Free;
  end;

end;

function TTwitterBot2.DoUploadMediaStatus_v1(MediaID: string; WaitSeconds: integer): boolean;
var
  HttpClient: THttpClient;
  Response: TmemoryStream;
  Params: string;
  WSeconds: integer;
  Json: string;
begin
  result := false;

  WSeconds := WaitSeconds;
  HttpClient := CreateHttpClient;
  Response := TMemoryStream.Create;
  try
    try
      HttpClient.ContentType := 'application/json';

      Params := '';
      AddQueryStr(Params, 'command', 'STATUS');
      AddQueryStr(Params, 'media_id', MediaID);
      HttpClient.CustHeaders.Add('Authorization', CreateSignature(hmGet, UploadMediaURL_v1, Params));
      var StartTime := now;
      while SecondsBetween(StartTime, Now) < 30 do begin // max wait time = 30 sec
        if Assigned(LogQueue) then
          LogQueue.Add('status=waiting ' + WSeconds.ToString + ' seconds');
        if not SleepNow(WSeconds) then exit;

        Response.Clear;
        HttpClient.Get(UploadMediaURL_v1 + '?command=STATUS&media_id=' + MediaID, Response);

        Json := MemoryStreamToString(Response);
        var State := GetJsonValueByName(Json, 'processing_info.state');
        if State = 'succeeded' then begin
          if Assigned(LogQueue) then
            LogQueue.Add('DoUploadMediaStatus_v1: Done');
          result := true;
          exit;
        end else if State = 'failed' then begin
          if Assigned(LogQueue) then
            LogQueue.Add('DoUploadMediaStatus_v1: failed');
          exit;
        end else begin // State = 'in_progress'
          WSeconds := GetWaitSecond(Json);
        end;
      end;
      if Assigned(LogQueue) then
        LogQueue.Add('Timeout');
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoUploadMediaStatus_v1: ' + e.Message);
    end;
  finally
    HttpClient.Free;
    Response.Free;
  end;
end;



// ---------------------------------------------
//     v2
// ---------------------------------------------

procedure TTwitterBot2.Authorize_v2;
  function GetRandomString(ALength: integer): string;
  const
    ACharSequence = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890';
  var
    Len: integer;
  begin
    Len := Length(ACharSequence);
    SetLength(result, ALength);
    Randomize;
    for var i := Low(result) to High(result) do
      result[i] := ACharSequence.Chars[Random(Len)]
  end;
  function GetBase64_SHA256(Text: string): string;
  var
    Bytes: TBytes;
  begin
    Bytes := THashSHA2.GetHashBytes(Text, SHA256);
    result := TNetEncoding.Base64.EncodeBytesToString(Bytes);
    result := result.Split(['='])[0];
    result := result.Replace('+', '-');
    result := result.Replace('/', '_');

  end;
var
  Params, URL: string;
begin
  try
    v2_State := GetRandomString(16);
    v2_CodeVerifier := GetRandomString(64);
    if v2_ChallengeMethod = 'plain' then
      v2_Challenge := v2_CodeVerifier
    else
      v2_Challenge := GetBase64_SHA256(v2_CodeVerifier);

    FHTTPProc := THTTPProc.hpAuth_v2;

    Params := '';
    AddQueryStr(Params, 'response_type', 'code');
    AddQueryStr(Params, 'client_id', v2_ClientId);
    AddQueryStr(Params, 'redirect_uri', v2_RedirectURL);
    AddQueryStr(Params, 'scope', UrlEncode(v2_Scopes));
    AddQueryStr(Params, 'state', v2_State);
    AddQueryStr(Params, 'code_challenge', v2_Challenge);
    AddQueryStr(Params, 'code_challenge_method', v2_ChallengeMethod);
    URL := AuthorizeURL + '?' + Params;

    if not URL.IsEmpty then begin
      v2_AuthState := TAuthState.osInProcess;
      if Assigned(FAuthStateChanged) then
        FAuthStateChanged(v2_AuthState, TAPIVersion.av20);

      if Assigned(FBrowserFormShow) then // open Browser
        FBrowserFormShow(URL, 'Login and authorize a app');
    end;
  except
    on e: exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TTwitterBot2.Authorize_v2: ' + e.Message);
  end;
end;

// get Access tokens after your user permitted your app
function TTwitterBot2.GetAccessToken_v2(Url: string): string;
var
  URI: TURI;
  HTTPClient: THTTPClient;
  Body: TStringList;
  Stream: TMemoryStream;
begin
  result := '';

  if Assigned(FBrowserFormShow) then  // close Browser
    FBrowserFormShow('', '');

  HTTPClient := CreateHttpClient;
  Body := TStringList.Create;
  Stream := TMemoryStream.Create;
  try
    try
      URI.Create(URL);

      // Basic Auth
      HTTPClient.PreemptiveAuthentication := true;
      HTTPClient.CredentialsStorage.AddCredential(TCredentialsStorage.TCredential.Create(TAuthTargetType.Server,
          '', '', v2_ClientID, v2_ClientSecret));
      HTTPClient.ContentType := 'application/x-www-form-urlencoded';

      Body.Add('code=' + Uri.ParameterByName['code']);
      Body.Add('grant_type=authorization_code');
      Body.Add('client_id=' + v2_ClientID);
      Body.Add('redirect_uri=' + v2_RedirectUrl);
      Body.Add('code_verifier=' + v2_CodeVerifier);

      HTTPClient.Post(GetAccessTokenURL, Body, Stream);

      result := MemoryStreamToString(Stream);
      // result := Response.ContentAsString(TEncoding.UTF8);
      if result.Contains('refresh_token') and result.Contains('access_token') then begin
        if ParseTokensFromOrgJson(result, Now) then begin
          v2_AuthState := TAuthState.osAuthorized;
          if Assigned(FAuthStateChanged) then
            FAuthStateChanged(v2_AuthState, TAPIVersion.av20);
          if Assigned(FSetJsonEvent_v2) then
            FSetJsonEvent_v2(Get_v2_TokensAsJson);
          TFile.WriteAllText('Tokens_v2.txt', result, TEncoding.UTF8);
        end;
      end else
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoGetAccessToken_v2: ' + result);

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoGetAccessToken_v2: ' + e.Message);
    end;
  finally
    HTTPClient.Free;
    Body.Free;
    Stream.Free;
  end;

end;

// {"token_type":"bearer",
// "expires_in":7200,
// "access_token":"VlR...",
// "scope":"tweet.write tweet.read offline.access",
// "refresh_token":"Wm5..."}

function TTwitterBot2.ParseTokensFromOrgJson(var Json: string; DateTime: TDateTime): boolean;
var
  JSONObject: TJSONObject;
  JsonValue: TJsonValue;
begin
  result := false;

  if not Json.Contains('refresh_token') then
    exit;

  JSONObject := nil;
  if not Json.IsEmpty then begin
    JSONObject := TJSONObject.ParseJSONValue(Json) as TJSONObject;
    if JSONObject <> nil then begin
      try
        if JSONObject.TryGetValue('access_token', JsonValue) then
          v2_BearerToken := JsonValue.Value;
        if JSONObject.TryGetValue('refresh_token', JsonValue) then
          v2_RefreshToken := JsonValue.Value;
        if JSONObject.TryGetValue('expires_in', JsonValue) then
          v2_RefreshedAt := now; // 7200
        Json := JSONObject.ToString;
        result := true;
      finally
        JSONObject.Free;
      end;
    end;
  end;

end;

// refresh tokens after a bearer token expired
function TTwitterBot2.DoRefreshTokens_v2: string;
var
  HTTPClient: THTTPClient;
  Body: TStringList;
  Stream: TMemoryStream;
begin
  result := '';

  HTTPClient := CreateHttpClient;
  Body := TStringList.Create;
  Stream := TMemoryStream.Create;
  try
    try
      // Basic Auth
      HTTPClient.PreemptiveAuthentication := true;
      HTTPClient.CredentialsStorage.AddCredential(TCredentialsStorage.TCredential.Create(TAuthTargetType.Server,
          '', '', v2_ClientID, v2_ClientSecret));
      HTTPClient.ContentType := 'application/x-www-form-urlencoded';

      Body.Add('refresh_token=' + v2_RefreshToken);
      Body.Add('grant_type=refresh_token');
      Body.Add('client_id=' + v2_ClientID);

      HTTPClient.Post(GetAccessTokenURL, Body, Stream);

      result := MemoryStreamToString(Stream);
      if result.Contains('refresh_token') and result.Contains('access_token') then
        ParseTokensFromOrgJson(result, Now);

      if Assigned(LogQueue) then
        LogQueue.Add('TTwitterBot2.DoRefreshTokens_v2_2: Refreshed');

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoRefreshTokens_v2: ' + e.Message);
    end;
  finally
    HTTPClient.Free;
    Body.Free;
    Stream.Free;
  end;

end;

function TTwitterBot2.DoPostTweet_v2(Text: string; MediaID: string): string;
type
  TTweetMedia = record
    media_ids: TArray<string>;
  end;
  TTweet = record
    text: string;
  end;
  TTweetwMedia = record
    text: string;
    media: TTweetMedia;
  end;
var
  HTTPClient: THTTPClient;
  Response: IHTTPResponse;

  Body: TStringStream;
  Stream: TMemoryStream;

  Json: string;
  Serial: TJsonSerializer;
  TweetwMedia: TTweetwMedia;
  Tweet: TTweet;
  TweetMedia: TTweetMedia;
begin
  result := '';

  Json := '';
  Serial := TJsonSerializer.Create;
  try
    try
      if MediaID.IsEmpty then begin     // need ?
        Tweet.text := Text;
        Json := Serial.Serialize<TTweet>(Tweet);
      end else begin
        TweetwMedia.text := Text;
        TweetMedia.media_ids := [MediaID];
        TweetwMedia.media := TweetMedia;
        Json := Serial.Serialize<TTweetwMedia>(TweetwMedia);
      end;
    except

    end;
  finally
    Serial.Free;
  end;

  HTTPClient := CreateHttpClient;
  Body := TStringStream.Create(Json, TEncoding.UTF8);
  Stream := TMemoryStream.Create;
  try
    try
      HTTPClient.CustHeaders.Add('Authorization', 'Bearer ' + v2_BearerToken);
      HTTPClient.ContentType := 'application/json';
      HTTPClient.AcceptCharSet := 'utf-8';

      Body.Seek(0, TSeekOrigin.soBeginning);
      Response := HTTPClient.Post(PostTweetUrl_v2, Body, Stream);

      LogRateLimit(Response.Headers);

      result := MemoryStreamToString(Stream);

    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoPostTweet_v2_2: ' + e.Message);
    end;
  finally
    HTTPClient.Free;
    Body.Free;
    Stream.Free;
  end;

end;

function TTwitterBot2.DoGetAboutMe: string;
var
  HTTPClient: THTTPClient;
  Response: IHTTPResponse;
  Stream: TMemoryStream;
begin
  result := '';

  HTTPClient := CreateHttpClient;
  Stream := TMemoryStream.Create;
  try
    try
      HTTPClient.CustHeaders.Add('Authorization', 'Bearer ' + v2_BearerToken);
      HTTPClient.ContentType := 'application/json';
      HTTPClient.AcceptCharSet := 'utf-8';

      Response := HTTPClient.Get(AboutMeUrl, Stream);
      if Response.StatusCode = 200 then begin
        result := MemoryStreamToString(Stream);
        // result := Response.ContentAsString(TEncoding.UTF8);
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoGetAboutMe_2_result: ' + result);
      end;
    except
      on e: exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TTwitterBot2.DoGetAboutMe_2: ' + e.Message);
    end;
  finally
    HTTPClient.Free;
    Stream.Free;
  end;
end;

end.
